/*
 *  Copyright (C) 2006 José María Cañas Plaza
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/**
 * jdec simulated3D driver provides video images to color variables from simulated3D
 * cameras using simulated3D library.
 *
 * This 4.3 version includes support for variable images. 
 *
 * @file simulated3D.c
 * @author Sara Marugán Alonso <s.marugan@alumnos.urjc.es>
 * @version 4.3
 * @date 2009-02-27
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include "progeo.h"
#include <forms.h>
#include "jde.h"
#include "simulated3D.h"
#include "xformsutils.h"
#include <interfaces/varcolor.h>

/** Max number of cameras detected by simulated3D driver.*/
#define MAXCAM 8
/** Image standard number of rows*/
#define SIFNTSC_ROWS 240
/** Image standard number of columns*/
#define SIFNTSC_COLUMNS 320

/** pthread identifiers for jdec simulated3D driver threads.*/
pthread_t simulated3D_th[MAXCAM];
/** Arguments for simulated3D threads*/
int args[MAXCAM];
/** pthread state variable for jdec simulated3D driver.*/
int state[MAXCAM];
/** pthread mutex for jdec simulated3D driver.*/
pthread_mutex_t mymutex[MAXCAM];
/** pthread condition variable for jdec simulated3D driver.*/
pthread_cond_t condition[MAXCAM];

/** variable to detect when the pthread is created.*/
int simulated3D_thread_created[MAXCAM];
/** variable to detect when simulated3D driver was cleaned up.*/
int simulated3D_cleaned_up=0;
/** variable to detect when simulated3D driver was setup.*/
int simulated3D_setup=0;
/** variable to detect when pthread must end execution.*/
int simulated3D_terminate_command=0;

/* simulated3D driver API options */
/** simulated3D driver name.*/
char driver_name[256]="simulated3D";
/** colors detected in config file.*/
int serve_color[MAXCAM];
/** structure to know what colors are active in the gui.*/
int color_active[MAXCAM];

/** width of each served video**/
int width[MAXCAM];
/** height of each served video**/
int height[MAXCAM];
/** id set to colorA schema.*/
int colorA_schema_id;
/** id set to colorB schema.*/
int colorB_schema_id;
/** id set to colorC schema.*/
int colorC_schema_id;
/** id set to colorD schema.*/
int colorD_schema_id;
/** id for varcolorA schema.*/
int varcolorA_schema_id;
/** id for varcolorB schema.*/
int varcolorB_schema_id;
/** id for varcolorC schema.*/
int varcolorC_schema_id;
/** id for varcolorD schema.*/
int varcolorD_schema_id;

/*API variables servidas*/
/** 'colorA' schema image data*/
char *colorA; /* sifntsc image itself */
/** 'colorA' schema clock*/
unsigned long int imageA_clock;

/** 'colorB' schema image data*/
char *colorB; /* sifntsc image itself */
/** 'colorB' schema clock*/
unsigned long int imageB_clock;

/** 'colorC' schema image data*/
char *colorC; /* sifntsc image itself */
/** 'colorC' schema clock*/
unsigned long int imageC_clock;

/** 'colorD' schema image data*/
char *colorD; /* sifntsc image itself */
/** 'colorD' schema clock*/
unsigned long int imageD_clock;

Varcolor myA,myB,myC,myD;


/*Contadores de referencias*/
/** colorA ref counter*/
int colorA_refs=0;
/** colorB ref counter*/
int colorB_refs=0;
/** colorC ref counter*/
int colorC_refs=0;
/** colorD ref counter*/
int colorD_refs=0;
/** varcolorA ref counter*/
int varcolorA_refs=0;
/** varcolorB ref counter*/
int varcolorB_refs=0;
/** varcolorC ref counter*/
int varcolorC_refs=0;
/** varcolorD ref counter*/
int varcolorD_refs=0;

/* camera configfiles */
char cameraINfileA[MAX_BUFFER];
char cameraINfileB[MAX_BUFFER];
char cameraINfileC[MAX_BUFFER];
char cameraINfileD[MAX_BUFFER];

/** mutex for ref counters*/
pthread_mutex_t refmutex;


/******************************************** simulated variables ***************************************/
/* world file */
char worldfile[512];

/* color */
char readyColorAA[SIFNTSC_ROWS*SIFNTSC_COLUMNS*3]; char notReadyColorAA[SIFNTSC_ROWS*SIFNTSC_COLUMNS*3];
char readyColorBB[SIFNTSC_ROWS*SIFNTSC_COLUMNS*3]; char notReadyColorBB[SIFNTSC_ROWS*SIFNTSC_COLUMNS*3];
char readyColorCC[SIFNTSC_ROWS*SIFNTSC_COLUMNS*3]; char notReadyColorCC[SIFNTSC_ROWS*SIFNTSC_COLUMNS*3];
char readyColorDD[SIFNTSC_ROWS*SIFNTSC_COLUMNS*3]; char notReadyColorDD[SIFNTSC_ROWS*SIFNTSC_COLUMNS*3];
char *notReadyColorSimA; char *notReadyColorSimB; char *notReadyColorSimC; char *notReadyColorSimD;
char *ReadyColorSimA; char *ReadyColorSimB; char *ReadyColorSimC; char *ReadyColorSimD;

/* varcolor */
char *varnotReadyColorAA; char *varnotReadyColorBB; char *varnotReadyColorCC; char *varnotReadyColorDD;
char *varReadyColorAA; char *varReadyColorBB; char *varReadyColorCC; char *varReadyColorDD;
char *varnotReadyColorSimA; char *varnotReadyColorSimB; char *varnotReadyColorSimC; char *varnotReadyColorSimD;
char *varReadyColorSimA; char *varReadyColorSimB; char *varReadyColorSimC; char *varReadyColorSimD;

/* 3d objects */
float DELTA;
HPoint3D origin;
HPoint3D axisX[2], axisY[2], axisZ[2];
HPoint3D gridYZ[SLOTS*2*2], gridXZ[SLOTS*2*2], gridXY[SLOTS*2*2];
HPoint3D room[MAX_LINES_IN_ROOM*2]; /* mm */
int room_lines=0;

/* declarations for simulated3D cameras */
TPinHoleCamera cameras[MAXCAM];
/** number of simulated3D cameras detected.*/
int camCount=0;
/** int variable to detect when a image was captured.*/
unsigned long int lastimage=0;

/* default camera values */
float CAM1_X, CAM1_Y, CAM1_Z, CAM1_H, CAM1F_X, CAM1F_Y, CAM1F_Z, CAM1F_H, CAM1_FDIST, CAM1_U0, CAM1_V0, CAM1_ROLL;
float CAM2_X, CAM2_Y, CAM2_Z, CAM2_H, CAM2F_X, CAM2F_Y, CAM2F_Z, CAM2F_H, CAM2_FDIST, CAM2_U0, CAM2_V0, CAM2_ROLL;
float CAM3_X, CAM3_Y, CAM3_Z, CAM3_H, CAM3F_X, CAM3F_Y, CAM3F_Z, CAM3F_H, CAM3_FDIST, CAM3_U0, CAM3_V0, CAM3_ROLL;
float CAM4_X, CAM4_Y, CAM4_Z, CAM4_H, CAM4F_X, CAM4F_Y, CAM4F_Z, CAM4F_H, CAM4_FDIST, CAM4_U0, CAM4_V0, CAM4_ROLL;
float VCAM_X, VCAM_Y, VCAM_Z, VCAM_H, VCAMF_X, VCAMF_Y, VCAMF_Z, VCAMF_H, VCAM_FDIST, VCAM_U0, VCAM_V0, VCAM_ROLL;

/* image visualization and auxiliar 3d coordenates variables*/
int room_on=1;
int camaxis_on=0; 
int grid_on=0; 
int axis_on=0;

/* exported variables */
int simulator_cycle=10; /* ms */

/* ------------------------------------------------------------ VISUAL FUNCTIONS ---------------------------------------------------------------- */

int simulator_drawline(HPoint2D p1, HPoint2D p2, char sel, FL_COLOR colour, int image)
/* it takes care of important features: before/behind the focal plane, inside/outside the image frame */
{
  HPoint2D gooda,goodb;

  if(image==READYCOLOR){
    if(sel==0 && displayline(p1,p2,&gooda,&goodb,cameras[0])==1)
	lineinimage(notReadyColorSimA,(int)gooda.y,cameras[0].rows-1-(int)gooda.x,(int)goodb.y,cameras[0].rows-1-(int)goodb.x,colour,cameras[0].columns,cameras[0].rows);
    else if(sel==1 && displayline(p1,p2,&gooda,&goodb,cameras[1])==1) 		
	lineinimage(notReadyColorSimB,(int)gooda.y,cameras[1].rows-1-(int)gooda.x,(int)goodb.y,cameras[1].rows-1-(int)goodb.x,colour,cameras[1].columns,cameras[1].rows);
    else if(sel==2 && displayline(p1,p2,&gooda,&goodb,cameras[2])==1)		
	lineinimage(notReadyColorSimC,(int)gooda.y,cameras[2].rows-1-(int)gooda.x,(int)goodb.y,cameras[2].rows-1-(int)goodb.x,colour,cameras[2].columns,cameras[2].rows);
    else if(sel==3 && displayline(p1,p2,&gooda,&goodb,cameras[3])==1)
 	lineinimage(notReadyColorSimD,(int)gooda.y,cameras[3].rows-1-(int)gooda.x,(int)goodb.y,cameras[3].rows-1-(int)goodb.x,colour,cameras[3].columns,cameras[3].rows);

    else if(sel==4 && displayline(p1,p2,&gooda,&goodb,cameras[4])==1)
	lineinimage(varnotReadyColorSimA,(int)gooda.y,cameras[4].rows-1-(int)gooda.x,(int)goodb.y,cameras[4].rows-1-(int)goodb.x,colour,cameras[4].columns,cameras[4].rows);
    else if(sel==5 && displayline(p1,p2,&gooda,&goodb,cameras[5])==1) 		
	lineinimage(varnotReadyColorSimB,(int)gooda.y,cameras[5].rows-1-(int)gooda.x,(int)goodb.y,cameras[5].rows-1-(int)goodb.x,colour,cameras[5].columns,cameras[5].rows);
    else if(sel==6 && displayline(p1,p2,&gooda,&goodb,cameras[6])==1) 		
	lineinimage(varnotReadyColorSimC,(int)gooda.y,cameras[6].rows-1-(int)gooda.x,(int)goodb.y,cameras[6].rows-1-(int)goodb.x,colour,cameras[6].columns,cameras[6].rows);
    else if(sel==7 && displayline(p1,p2,&gooda,&goodb,cameras[7])==1)
 	lineinimage(varnotReadyColorSimD,(int)gooda.y,cameras[7].rows-1-(int)gooda.x,(int)goodb.y,cameras[7].rows-1-(int)goodb.x,colour,cameras[7].columns,cameras[7].rows);
  };
 return 0;
}


/* virtual object generator */
void object_generator(int index)
{
  int i;
  HPoint2D a,b;
  char *auxsim;
  
	    /*white background*/
	    switch(index){
		case 0: 
		    	reset_buffer(notReadyColorSimA,(double)255,(double)255,(double)255,cameras[0].columns,cameras[0].rows);
			break;
		case 1: 
		    	reset_buffer(notReadyColorSimB,(double)255,(double)255,(double)255,cameras[1].columns,cameras[1].rows);
			break;
		case 2: 
		    	reset_buffer(notReadyColorSimC,(double)255,(double)255,(double)255,cameras[2].columns,cameras[2].rows);
			break;
		case 3: 
		    	reset_buffer(notReadyColorSimD,(double)255,(double)255,(double)255,cameras[3].columns,cameras[3].rows);
			break;
		case 4: 
		    	reset_buffer(varnotReadyColorSimA,(double)255,(double)255,(double)255,cameras[4].columns,cameras[4].rows);
			break;
		case 5: 
		    	reset_buffer(varnotReadyColorSimB,(double)255,(double)255,(double)255,cameras[5].columns,cameras[5].rows);
			break;
		case 6: 
		    	reset_buffer(varnotReadyColorSimC,(double)255,(double)255,(double)255,cameras[6].columns,cameras[6].rows);
			break;
		case 7: 
		    	reset_buffer(varnotReadyColorSimD,(double)255,(double)255,(double)255,cameras[7].columns,cameras[7].rows);
			break;
	    }
	    
	    /*grid in image*/
  	    if(grid_on==1){
	      for(i=0;i<SLOTS;i++){
		project(gridXY[2*i],&a,cameras[index]);
		project(gridXY[2*i+1],&b,cameras[index]);
		simulator_drawline(a,b,index,FL_RED,READYCOLOR);
		project(gridXY[2*SLOTS+2*i],&a,cameras[index]);
		project(gridXY[2*SLOTS+2*i+1],&b,cameras[index]);
		simulator_drawline(a,b,index,FL_RED,READYCOLOR);
	      }
	      
	      for(i=0;i<SLOTS;i++){
		project(gridYZ[2*i],&a,cameras[index]);
		project(gridYZ[2*i+1],&b,cameras[index]);
		simulator_drawline(a,b,index,FL_PALEGREEN,READYCOLOR);
		project(gridYZ[2*SLOTS+2*i],&a,cameras[index]);
		project(gridYZ[2*SLOTS+2*i+1],&b,cameras[index]);
		simulator_drawline(a,b,index,FL_PALEGREEN,READYCOLOR);
	      }
		
	      for(i=0;i<SLOTS;i++){
		project(gridXZ[2*i],&a,cameras[index]);
		project(gridXZ[2*i+1],&b,cameras[index]);
		simulator_drawline(a,b,index,FL_BLUE,READYCOLOR);
		project(gridXZ[2*SLOTS+2*i],&a,cameras[index]);
		project(gridXZ[2*SLOTS+2*i+1],&b,cameras[index]);
		simulator_drawline(a,b,index,FL_BLUE,READYCOLOR);
	      }				 	
	    }
		      
	    /* drawing of X-Y-Z axis*/
	    if ((axis_on==1)||(grid_on==1)){
		project(axisX[0],&a,cameras[index]);
		project(axisX[1],&b,cameras[index]);
		simulator_drawline(a,b,index,FL_PALEGREEN,READYCOLOR);
	
		project(axisY[0],&a,cameras[index]);
		project(axisY[1],&b,cameras[index]);
		simulator_drawline(a,b,index,FL_BLUE,READYCOLOR);
	
		project(axisZ[0],&a,cameras[index]);
		project(axisZ[1],&b,cameras[index]);
		simulator_drawline(a,b,index,FL_RED,READYCOLOR);
	    }
	    
	    /* camera2, camera3, camera4 and virtual camera seen in camera */
	    if (camaxis_on==1){
	        project(cameras[1].position,&a,cameras[index]);
	        project(cameras[1].foa,&b,cameras[index]);
	        simulator_drawline(a,b,index,FL_WHEAT,READYCOLOR);      
	        project(cameras[2].position,&a,cameras[index]);
	        project(cameras[2].foa,&b,cameras[index]);
	        simulator_drawline(a,b,index,FL_WHEAT,READYCOLOR);      
	        project(cameras[3].position,&a,cameras[index]);
	        project(cameras[3].foa,&b,cameras[index]);
	        simulator_drawline(a,b,index,FL_WHEAT,READYCOLOR);
	        project(cameras[4].position,&a,cameras[index]);
	        project(cameras[4].foa,&b,cameras[index]);
	        simulator_drawline(a,b,index,FL_WHEAT,READYCOLOR);
	    }
	    
	    /* room in camera */
	    if(room_on==1){
	      for(i=0;i<room_lines;i++){
		project(room[i*2+0],&a,cameras[index]);
		project(room[i*2+1],&b,cameras[index]);
		simulator_drawline(a,b,index,FL_BLACK,READYCOLOR);
	      }
	    }
	    
	    switch(index){
		case 0: 
		    	auxsim=notReadyColorSimA; notReadyColorSimA=ReadyColorSimA; ReadyColorSimA=auxsim;
			break;
		case 1: 
		    	auxsim=notReadyColorSimB; notReadyColorSimB=ReadyColorSimB; ReadyColorSimB=auxsim;
			break;
		case 2: 
		    	auxsim=notReadyColorSimC; notReadyColorSimC=ReadyColorSimC; ReadyColorSimC=auxsim;
			break;
		case 3: 
		    	auxsim=notReadyColorSimD; notReadyColorSimD=ReadyColorSimD; ReadyColorSimD=auxsim;
			break;
		case 4: 
		    	auxsim=varnotReadyColorSimA; varnotReadyColorSimA=varReadyColorSimA; varReadyColorSimA=auxsim;
			break;
		case 5: 
		    	auxsim=varnotReadyColorSimB; varnotReadyColorSimB=varReadyColorSimB; varReadyColorSimB=auxsim;
			break;
		case 6: 
		    	auxsim=varnotReadyColorSimC; varnotReadyColorSimC=varReadyColorSimC; varReadyColorSimC=auxsim;
			break;
		case 7: 
		    	auxsim=varnotReadyColorSimD; varnotReadyColorSimD=varReadyColorSimD; varReadyColorSimD=auxsim;
			break;
  	   }
}


/* ------------------------------------------------------------- AUXILIAR SCHEMA FUNCTIONS ------------------------------------------------------------- */

/* gets the calibration of the camera from a file */
int load_cam_line(FILE *myfile,int cam)
{		
  char word1[MAX_BUFFER],word2[MAX_BUFFER];
  int i=0;
  char buffer_file[MAX_BUFFER];   

  buffer_file[0]=fgetc(myfile);
  if (feof(myfile)) return EOF;
  if (buffer_file[0]==(char)255) return EOF; 
  if (buffer_file[0]=='#') {while(fgetc(myfile)!='\n'); return 0;}
  if (buffer_file[0]==' ') {while(buffer_file[0]==' ') buffer_file[0]=fgetc(myfile);}
  if (buffer_file[0]=='\t') {while(buffer_file[0]=='\t') buffer_file[0]=fgetc(myfile);}

  /* Captures a line and then we will process it with sscanf checking that the last character is \n. We can't doit with fscanf because this function does not difference \n from blank space. */
  while((buffer_file[i]!='\n') && 
	(buffer_file[i] != (char)255) &&  
	(i<MAX_BUFFER-1) ) {
    buffer_file[++i]=fgetc(myfile);
  }
  
  if (i >= MAX_BUFFER-1) { 
    printf("%s...\n", buffer_file); 
    printf ("Line too long in config file!\n"); 
    exit(-1);
  }
  buffer_file[++i]='\0';


  if (sscanf(buffer_file,"%s",word1)!=1) return 0; 
  /* return EOF; empty line*/
  else {
    if (strcmp(word1,"positionX")==0){
      sscanf(buffer_file,"%s %s",word1,word2);
      cameras[cam].position.X=(float)atof(word2);
    }
    else if (strcmp(word1,"positionY")==0){
      sscanf(buffer_file,"%s %s",word1,word2);
      cameras[cam].position.Y=(float)atof(word2);
    }
    else if (strcmp(word1,"positionZ")==0){
      sscanf(buffer_file,"%s %s",word1,word2);
      cameras[cam].position.Z=(float)atof(word2);
    }
    else if (strcmp(word1,"positionH")==0){
      sscanf(buffer_file,"%s %s",word1,word2);
      cameras[cam].position.H=(float)atof(word2);
    }
    else if (strcmp(word1,"FOApositionX")==0){
      sscanf(buffer_file,"%s %s",word1,word2);
      cameras[cam].foa.X=(float)atof(word2);
    }
    else if (strcmp(word1,"FOApositionY")==0){
      sscanf(buffer_file,"%s %s",word1,word2);
      cameras[cam].foa.Y=(float)atof(word2);
    }
    else if (strcmp(word1,"FOApositionZ")==0){
      sscanf(buffer_file,"%s %s",word1,word2);
      cameras[cam].foa.Z=(float)atof(word2);
    }
    else if (strcmp(word1,"FOApositionH")==0){
      sscanf(buffer_file,"%s %s",word1,word2);
      cameras[cam].foa.H=(float)atof(word2);
    }
    else if (strcmp(word1,"roll")==0){
      sscanf(buffer_file,"%s %s",word1,word2);
      cameras[cam].roll=(float)atof(word2);
    }
    else if (strcmp(word1,"f")==0){
      sscanf(buffer_file,"%s %s",word1,word2);
      cameras[cam].fdistx=(float)atof(word2);
      cameras[cam].fdisty=(float)atof(word2);
     }
    else if (strcmp(word1,"fx")==0){
      sscanf(buffer_file,"%s %s",word1,word2);
      cameras[cam].fdistx=(float)atof(word2);
    }
    else if (strcmp(word1,"fy")==0){
      sscanf(buffer_file,"%s %s",word1,word2);
      cameras[cam].fdisty=(float)atof(word2);
     }
    else if (strcmp(word1,"skew")==0){
      sscanf(buffer_file,"%s %s",word1,word2);
      cameras[cam].skew=(float)atof(word2);
     }
    else if (strcmp(word1,"u0")==0){
      sscanf(buffer_file,"%s %s",word1,word2);
      cameras[cam].u0=(float)atof(word2);
    }
    else if (strcmp(word1,"v0")==0){
      sscanf(buffer_file,"%s %s",word1,word2);
      cameras[cam].v0=(float)atof(word2);
    } 
  }
 return 1;
}

/* gets the calibration of the camera from a file */
void load_cam(char *fich_in,int cam)
{
  FILE *entrada;
  int i;

  entrada=fopen(fich_in,"r");
   if(entrada==NULL){
     printf("simulated3D: camera input calibration file %s does not exits\n",fich_in);
   }else{
     do{i=load_cam_line(entrada,cam);}while(i!=EOF);
   } 
  fclose(entrada);
  update_camera_matrix(&cameras[cam]);
  /*mydebug();*/
}

/**
 * It reads a single line from config file, parses it and do the right thing.
 * @param myfile The config file descriptor.
 * @returns EOF in detects end of such file. Otherwise returns 0.
 */
int load_worldline(FILE *myfile)
{		
  char word1[MAX_BUFFER],word2[MAX_BUFFER],word3[MAX_BUFFER],word4[MAX_BUFFER],word5[MAX_BUFFER];
  char word6[MAX_BUFFER],word7[MAX_BUFFER],word8[MAX_BUFFER];
  char word[MAX_BUFFER];
  int i=0;
  char buffer_file[MAX_BUFFER];  
  char id; 
  char cameraINfile[MAX_BUFFER];

  buffer_file[0]=fgetc(myfile);
  if (feof(myfile)) return EOF;
  if (buffer_file[0]==(char)255) return EOF; 
  if (buffer_file[0]=='#') {while(fgetc(myfile)!='\n'); return 0;}
  if (buffer_file[0]==' ') {while(buffer_file[0]==' ') buffer_file[0]=fgetc(myfile);}
  if (buffer_file[0]=='\t') {while(buffer_file[0]=='\t') buffer_file[0]=fgetc(myfile);}

  /* Captures a line and then we will process it with sscanf checking that the last character is \n. We can't doit with fscanf because this function does not difference \n from blank space. */
  while((buffer_file[i]!='\n') && 
	(buffer_file[i] != (char)255) &&  
	(i<MAX_BUFFER-1) ) {
    buffer_file[++i]=fgetc(myfile);
  }
  
  if (i >= MAX_BUFFER-1) { 
    printf("%s...\n", buffer_file); 
    printf ("Line too long in config file!\n"); 
    exit(-1);
  }
  buffer_file[++i]='\0';


  if (sscanf(buffer_file,"%s",word)!=1) return 0; 
  /* return EOF; empty line*/
  else {
     if(strcmp(word,"worldline")==0){
	sscanf(buffer_file,"%s %s %s %s %s %s %s %s %s",word,word1,word2,word3,word4,word5,word6,word7,word8);
	room[room_lines*2+0].X=(float)atof(word1); room[room_lines*2+0].Y=(float)atof(word2); room[room_lines*2+0].Z=(float)atof(word3); room[room_lines*2+0].H=(float)atof(word4);
	room[room_lines*2+1].X=(float)atof(word5); room[room_lines*2+1].Y=(float)atof(word6); room[room_lines*2+1].Z=(float)atof(word7); room[room_lines*2+1].H=(float)atof(word8);
	room_lines++;
     }
     else if (strcmp(word,"cameraINfile")==0){
       sscanf(buffer_file,"%s %c %s",word,&id,cameraINfile);
       switch(id){
       	  case 'A': strcpy(cameraINfileA,cameraINfile);
                    break;
       	  case 'B': strcpy(cameraINfileB,cameraINfile);
                    break;
       	  case 'C': strcpy(cameraINfileC,cameraINfile);
                    break;
       	  case 'D': strcpy(cameraINfileD,cameraINfile);
	            break;
       }
     }
     else if (strcmp(word,"maxX")==0){
     }
     else if (strcmp(word,"minX")==0){
     }
     else if (strcmp(word,"maxY")==0){
     }
     else if (strcmp(word,"minY")==0){
     }
     else if (strcmp(word,"maxZ")==0){
     }
     else if (strcmp(word,"minZ")==0){
     }
  }
  return 1;
}

/* initializes all 3D objects */
int tridimensional_objects_init()
{
  int i=0;
  FILE *camroomconfig;
  printf("Creating 3d objects...\n");

  /* points and lines in the 3D world */
  /* axis in 3D */
  axisX[0].X=0;
  axisX[0].Y=0;
  axisX[0].Z=0;
  axisX[0].H=1.;
  axisX[1].X=AXIS_Z;
  axisX[1].Y=0;
  axisX[1].Z=0;
  axisX[1].H=1.;
  
  axisY[0].X=0;
  axisY[0].Y=0;
  axisY[0].Z=0;
  axisY[0].H=1.;
  axisY[1].X=0.;
  axisY[1].Y=AXIS_Z;
  axisY[1].Z=0;
  axisY[1].H=1.;
  
  axisZ[0].X=0;
  axisZ[0].Y=0;
  axisZ[0].Z=0;
  axisZ[0].H=1.;
  axisZ[1].X=0.;
  axisZ[1].Y=0.;
  axisZ[1].Z=AXIS_Z;
  axisZ[1].H=1.;
  
  DELTA=AXIS_Z/SLOTS;
  /* XYgrid, XZgrid, YZgrid in 3D */
  for(i=0;i<SLOTS;i++){     
    gridXY[2*i].X=DELTA*(i+1); 
    gridXY[2*i].Y=0; 
    gridXY[2*i].Z=0; 
    gridXY[2*i].H=1;
    gridXY[2*i+1].X=DELTA*(i+1); 
    gridXY[2*i+1].Y=AXIS_Z; 
    gridXY[2*i+1].Z=0; 
    gridXY[2*i+1].H=1;
    
    gridXY[2*SLOTS+2*i].X=0; 
    gridXY[2*SLOTS+2*i].Y=DELTA*(i+1);
    gridXY[2*SLOTS+2*i].Z=0;
    gridXY[2*SLOTS+2*i].H=1;
    gridXY[2*SLOTS+2*i+1].X=AXIS_Z;
    gridXY[2*SLOTS+2*i+1].Y=DELTA*(i+1);
    gridXY[2*SLOTS+2*i+1].Z=0;
    gridXY[2*SLOTS+2*i+1].H=1;
    
    gridYZ[2*i].X=0;
    gridYZ[2*i].Y=DELTA*(i+1);
    gridYZ[2*i].Z=0;
    gridYZ[2*i].H=1;
    gridYZ[2*i+1].X=0;
    gridYZ[2*i+1].Y=DELTA*(i+1);
    gridYZ[2*i+1].Z=AXIS_Z;
    gridYZ[2*i+1].H=1;
    
    gridYZ[2*SLOTS+2*i].X=0;
    gridYZ[2*SLOTS+2*i].Y=0;
    gridYZ[2*SLOTS+2*i].Z=DELTA*(i+1);
    gridYZ[2*SLOTS+2*i].H=1;
    gridYZ[2*SLOTS+2*i+1].X=0;
    gridYZ[2*SLOTS+2*i+1].Y=AXIS_Z;
    gridYZ[2*SLOTS+2*i+1].Z=DELTA*(i+1);
    gridYZ[2*SLOTS+2*i+1].H=1;
    
    gridXZ[2*i].X=DELTA*(i+1);
    gridXZ[2*i].Y=0;
    gridXZ[2*i].Z=0;
    gridXZ[2*i].H=1;
    gridXZ[2*i+1].X=DELTA*(i+1);
    gridXZ[2*i+1].Y=0;
    gridXZ[2*i+1].Z=AXIS_Z;
    gridXZ[2*i+1].H=1;
    
    gridXZ[2*SLOTS+2*i].X=0;
    gridXZ[2*SLOTS+2*i].Y=0;
    gridXZ[2*SLOTS+2*i].Z=DELTA*(i+1);
    gridXZ[2*SLOTS+2*i].H=1;
    gridXZ[2*SLOTS+2*i+1].X=AXIS_Z;
    gridXZ[2*SLOTS+2*i+1].Y=0;
    gridXZ[2*SLOTS+2*i+1].Z=DELTA*(i+1);
    gridXZ[2*SLOTS+2*i+1].H=1;
  }

  camroomconfig=fopen(worldfile,"r");
  if(camroomconfig==NULL){
    return -1;
  }else{
    printf("file %s exists. reading data...\n",worldfile);
    do{i=load_worldline(camroomconfig);}while(i!=EOF);
  }

  if(serve_color[0]) load_cam(cameraINfileA,0);
  if(serve_color[1]) load_cam(cameraINfileB,1);
  if(serve_color[2]) load_cam(cameraINfileC,2);
  if(serve_color[3]) load_cam(cameraINfileD,3);
  if(serve_color[4]) load_cam(cameraINfileA,4);
  if(serve_color[5]) load_cam(cameraINfileB,5);
  if(serve_color[6]) load_cam(cameraINfileC,6);
  if(serve_color[7]) load_cam(cameraINfileD,7);
 
  printf("3d objects created\n");
  return 0;
}
/*--fin 3d init-**/



/************************************** simulated3D DRIVER FUNCTIONS ****************************/

/** simulated3D driver closing function invoked when stopping driver.*/
void simulated3D_terminate(){
  simulated3D_terminate_command=1;
  free(varReadyColorAA); 
  free(varnotReadyColorAA);
  free(varReadyColorBB); 
  free(varnotReadyColorBB);
  free(varReadyColorCC); 
  free(varnotReadyColorCC);
  free(varReadyColorDD); 
  free(varnotReadyColorDD);

  printf("driver simulated3D off\n");
}

/** colorA run function following jdec platform API schemas.
 *  @param father Father id for this schema.
 *  @param brothers Brothers for this schema.
 *  @param fn arbitration function for this schema.
 *  @return integer resuming result.*/
int mycolorA_run(int father, int *brothers, arbitration fn){
   pthread_mutex_lock(&refmutex);
   if (colorA_refs>0){
      colorA_refs++;
      pthread_mutex_unlock(&refmutex);
   }
   else{
      colorA_refs=1;
      pthread_mutex_unlock(&refmutex);
      if((serve_color[0]==1)&&(color_active[0]==0)){
         color_active[0]=1;
         printf("colorA schema run (simulated3D driver)\n");
         all[colorA_schema_id].father = father;
         all[colorA_schema_id].fps = 0.;
         all[colorA_schema_id].k =0;
         all[colorA_schema_id].k =0;
         put_state(colorA_schema_id,winner);
         /* simulated3D thread goes winner */
         pthread_mutex_lock(&mymutex[0]);
         state[0]=winner;
         pthread_cond_signal(&condition[0]);
         pthread_mutex_unlock(&mymutex[0]);
      }
   }
   return 0;
}

/** colorA stop function following jdec platform API schemas.
 *  @return integer stoping result.*/
int mycolorA_stop(){

   pthread_mutex_lock(&refmutex);
   if (colorA_refs>1){
      colorA_refs--;
      pthread_mutex_unlock(&refmutex);
   }
   else{
      colorA_refs=0;
      pthread_mutex_unlock(&refmutex);
      if((serve_color[0]==1)&&(color_active[0]==1)){
         color_active[0]=0;
         put_state(colorA_schema_id,slept);
         printf("colorA schema stop (simulated3D driver)\n");
         /* simulated3D thread goes sleep */
         pthread_mutex_lock(&mymutex[0]);
         state[0]=slept;
         pthread_mutex_unlock(&mymutex[0]);
      }
   }
   return 0;
}

/** colorB run function following jdec platform API schemas.
 *  @param father Father id for this schema.
 *  @param brothers Brothers for this schema.
 *  @param fn arbitration function for this schema.
 *  @return integer resuming result.*/
int mycolorB_run(int father, int *brothers, arbitration fn){

   pthread_mutex_lock(&refmutex);
   if (colorB_refs>0){
      colorB_refs++;
      pthread_mutex_unlock(&refmutex);
   }
   else{
      colorB_refs=1;
      pthread_mutex_unlock(&refmutex);
      if((serve_color[1]==1)&&(color_active[1]==0)){
         color_active[1]=1;
         printf("colorB schema run (simulated3D driver)\n");
         all[colorB_schema_id].father = father;
         all[colorB_schema_id].fps = 0.;
         all[colorB_schema_id].k =0;
   
         put_state(colorB_schema_id,winner);

         /* simulated3D thread goes winner */
         pthread_mutex_lock(&mymutex[1]);
         state[1]=winner;
         pthread_cond_signal(&condition[1]);
         pthread_mutex_unlock(&mymutex[1]);
      }
   }
   return 0;
}

/** colorB stop function following jdec platform API schemas.
 *  @return integer stoping result.*/
int mycolorB_stop(){

   pthread_mutex_lock(&refmutex);
   if (colorB_refs>1){
      colorB_refs--;
      pthread_mutex_unlock(&refmutex);
   }
   else{
      colorB_refs=0;
      pthread_mutex_unlock(&refmutex);
      if((serve_color[1]==1)&&(color_active[1]==1)){
         color_active[1]=0;
         printf("colorB schema stop (simulated3D driver)\n");
         put_state(colorB_schema_id,slept);
         /* simulated3D thread goes sleep */
         pthread_mutex_lock(&mymutex[1]);
         state[1]=slept;
         pthread_mutex_unlock(&mymutex[1]);
      }
   }
   return 0;
}

/** colorC run function following jdec platform API schemas.
 *  @param father Father id for this schema.
 *  @param brothers Brothers for this schema.
 *  @param fn arbitration function for this schema.
 *  @return integer resuming result.*/
int mycolorC_run(int father, int *brothers, arbitration fn){

   pthread_mutex_lock(&refmutex);
   if (colorC_refs>0){
      colorC_refs++;
      pthread_mutex_unlock(&refmutex);
   }
   else{
      colorC_refs=1;
      pthread_mutex_unlock(&refmutex);
      if((serve_color[2]==1)&&(color_active[2]==0)){
         color_active[2]=1;
         printf("colorC schema run (simulated3D driver)\n");
         all[colorC_schema_id].father = father;
         all[colorC_schema_id].fps = 0.;
         all[colorC_schema_id].k =0;
         put_state(colorC_schema_id,winner);
         /* simulated3D thread goes winner */
         pthread_mutex_lock(&mymutex[2]);
         state[2]=winner;
         pthread_cond_signal(&condition[2]);
         pthread_mutex_unlock(&mymutex[2]);
      }
   }
   return 0;
}

/** colorC stop function following jdec platform API schemas.
 *  @return integer stoping result.*/
int mycolorC_stop(){
   
   pthread_mutex_lock(&refmutex);
   if (colorC_refs>1){
      colorC_refs--;
      pthread_mutex_unlock(&refmutex);
   }
   else{
      colorC_refs=0;
      pthread_mutex_unlock(&refmutex);
      if((serve_color[2]==1)&&(color_active[2]==1)){
         color_active[2]=0;
         printf("colorC schema stop (simulated3D driver)\n");
         put_state(colorC_schema_id,slept);
         /* simulated3D thread goes sleep */
         pthread_mutex_lock(&mymutex[2]);
         state[2]=slept;
         pthread_mutex_unlock(&mymutex[2]);
      }
   }
   return 0;
}

/** colorD run function following jdec platform API schemas.
 *  @param father Father id for this schema.
 *  @param brothers Brothers for this schema.
 *  @param fn arbitration function for this schema.
 *  @return integer resuming result.*/
int mycolorD_run(int father, int *brothers, arbitration fn){

   pthread_mutex_lock(&refmutex);
   if (colorD_refs>0){
      colorD_refs++;
      pthread_mutex_unlock(&refmutex);
   }
   else{
      colorD_refs=1;
      pthread_mutex_unlock(&refmutex);
      if((serve_color[3]==1)&&(color_active[3]==0)){
         color_active[3]=1;
         printf("colorD schema run (simulated3D driver)\n");
         all[colorD_schema_id].father = father;
         all[colorD_schema_id].fps = 0.;
         all[colorD_schema_id].k =0;
         put_state(colorD_schema_id,winner);
         /* simulated3D thread goes winner */
         pthread_mutex_lock(&mymutex[3]);
         state[3]=winner;
         pthread_cond_signal(&condition[3]);
         pthread_mutex_unlock(&mymutex[3]);
      }
   }
   return 0;
}

/** colorD stop function following jdec platform API schemas.
 *  @return integer stoping result.*/
int mycolorD_stop(){
   
   pthread_mutex_lock(&refmutex);
   if (colorD_refs>1){
      colorD_refs--;
      pthread_mutex_unlock(&refmutex);
   }
   else{
      colorD_refs=0;
      pthread_mutex_unlock(&refmutex);
      if((serve_color[3]==1)&&(color_active[3]==1)){
         color_active[3]=0;
         printf("colorD schema stop (simulated3D driver)\n");
         put_state(colorD_schema_id,slept);
         /* simulated3D thread goes sleep */
         pthread_mutex_lock(&mymutex[3]);
         state[3]=slept;
         pthread_mutex_unlock(&mymutex[3]);
      }
   }
   return 0;
}

/** varcolorA run function following jdec platform API schemas.
 *  @param father Father id for this schema.
 *  @param brothers Brothers for this schema.
 *  @param fn arbitration function for this schema.
 *  @return integer resuming result.*/
int myvarcolorA_run(int father, int *brothers, arbitration fn){
   pthread_mutex_lock(&refmutex);
   if (varcolorA_refs>0){
      varcolorA_refs++;
      pthread_mutex_unlock(&refmutex);
   }
   else{
      varcolorA_refs=1;
      pthread_mutex_unlock(&refmutex);
      if((serve_color[4]==1)&&(color_active[4]==0)){
         color_active[4]=1;
         printf("varcolorA schema run (simulated3D driver)\n");
         all[varcolorA_schema_id].father = father;
         all[varcolorA_schema_id].fps = 0.;
         all[varcolorA_schema_id].k =0;
         put_state(varcolorA_schema_id,winner);
         /* simulated3D thread goes winner */
         pthread_mutex_lock(&mymutex[4]);
         state[4]=winner;
         pthread_cond_signal(&condition[4]);
         pthread_mutex_unlock(&mymutex[4]);
      }
   }
   return 0;
}

/** varcolorA stop function following jdec platform API schemas.
 *  @return integer stoping result.*/
int myvarcolorA_stop(){
   pthread_mutex_lock(&refmutex);
   if (varcolorA_refs>1){
      varcolorA_refs--;
      pthread_mutex_unlock(&refmutex);
   }
   else{
      varcolorA_refs=0;
      pthread_mutex_unlock(&refmutex);
      if((serve_color[4]==1)&&(color_active[4]==1)){
         color_active[4]=0;
         printf("varcolorA schema stop (simulated3D driver)\n");
         put_state(varcolorA_schema_id,slept);
         /* simulated3D thread goes sleep */
         pthread_mutex_lock(&mymutex[4]);
         state[4]=slept;
         pthread_mutex_unlock(&mymutex[4]);
      }
   }
   return 0;
}

/** varcolorB run function following jdec platform API schemas.
 *  @param father Father id for this schema.
 *  @param brothers Brothers for this schema.
 *  @param fn arbitration function for this schema.
 *  @return integer resuming result.*/
int myvarcolorB_run(int father, int *brothers, arbitration fn){
   pthread_mutex_lock(&refmutex);
   if (varcolorB_refs>0){
      varcolorB_refs++;
      pthread_mutex_unlock(&refmutex);
   }
   else{
      varcolorB_refs=1;
      pthread_mutex_unlock(&refmutex);
      if((serve_color[5]==1)&&(color_active[5]==0)){
         color_active[5]=1;
         printf("varcolorB schema run (simulated3D driver)\n");
         all[varcolorB_schema_id].father = father;
         all[varcolorB_schema_id].fps = 0.;
         all[varcolorB_schema_id].k =0;
         put_state(varcolorB_schema_id,winner);
         /* simulated3D thread goes winner */
         pthread_mutex_lock(&mymutex[5]);
         state[5]=winner;
         pthread_cond_signal(&condition[5]);
         pthread_mutex_unlock(&mymutex[5]);
      }
   }
   return 0;
}

/** varcolorB stop function following jdec platform API schemas.
 *  @return integer stoping result.*/
int myvarcolorB_stop(){
   pthread_mutex_lock(&refmutex);
   if (varcolorB_refs>1){
      varcolorB_refs--;
      pthread_mutex_unlock(&refmutex);
   }
   else{
      varcolorB_refs=0;
      pthread_mutex_unlock(&refmutex);
      if((serve_color[5]==1)&&(color_active[5]==1)){
         color_active[5]=0;
         printf("varcolorB schema stop (simulated3D driver)\n");
         put_state(varcolorB_schema_id,slept);
         /* simulated3D thread goes sleep */
         pthread_mutex_lock(&mymutex[5]);
         state[5]=slept;
         pthread_mutex_unlock(&mymutex[5]);
      }
   }
   return 0;
}

/** varcolorC run function following jdec platform API schemas.
 *  @param father Father id for this schema.
 *  @param brothers Brothers for this schema.
 *  @param fn arbitration function for this schema.
 *  @return integer resuming result.*/
int myvarcolorC_run(int father, int *brothers, arbitration fn){
   pthread_mutex_lock(&refmutex);
   if (varcolorC_refs>0){
      varcolorC_refs++;
      pthread_mutex_unlock(&refmutex);
   }
   else{
      varcolorC_refs=1;
      pthread_mutex_unlock(&refmutex);
      if((serve_color[6]==1)&&(color_active[6]==0)){
         color_active[6]=1;
         printf("varcolorC schema run (simulated3D driver)\n");
         all[varcolorC_schema_id].father = father;
         all[varcolorC_schema_id].fps = 0.;
         all[varcolorC_schema_id].k =0;
         put_state(varcolorC_schema_id,winner);
         /* simulated3D thread goes winner */
         pthread_mutex_lock(&mymutex[6]);
         state[6]=winner;
         pthread_cond_signal(&condition[6]);
         pthread_mutex_unlock(&mymutex[6]);
      }
   }
   return 0;
}

/** varcolorC stop function following jdec platform API schemas.
 *  @return integer stoping result.*/
int myvarcolorC_stop(){
   pthread_mutex_lock(&refmutex);
   if (varcolorC_refs>1){
      varcolorC_refs--;
      pthread_mutex_unlock(&refmutex);
   }
   else{
      varcolorC_refs=0;
      pthread_mutex_unlock(&refmutex);
      if((serve_color[6]==1)&&(color_active[6]==1)){
         color_active[6]=0;
         printf("varcolorC schema stop (simulated3D driver)\n");
         put_state(varcolorC_schema_id,slept);
         /* simulated3D thread goes sleep */
         pthread_mutex_lock(&mymutex[6]);
         state[6]=slept;
         pthread_mutex_unlock(&mymutex[6]);
      }
   }
   return 0;
}

/** varcolorD run function following jdec platform API schemas.
 *  @param father Father id for this schema.
 *  @param brothers Brothers for this schema.
 *  @param fn arbitration function for this schema.
 *  @return integer resuming result.*/
int myvarcolorD_run(int father, int *brothers, arbitration fn){
   pthread_mutex_lock(&refmutex);
   if (varcolorD_refs>0){
      varcolorD_refs++;
      pthread_mutex_unlock(&refmutex);
   }
   else{
      varcolorD_refs=1;
      pthread_mutex_unlock(&refmutex);
      if((serve_color[7]==1)&&(color_active[7]==0)){
         color_active[7]=1;
         printf("varcolorD schema run (simulated3D driver)\n");
         all[varcolorD_schema_id].father = father;
         all[varcolorD_schema_id].fps = 0.;
         all[varcolorD_schema_id].k =0;
         put_state(varcolorD_schema_id,winner);
         /* simulated3D thread goes winner */
         pthread_mutex_lock(&mymutex[7]);
         state[7]=winner;
         pthread_cond_signal(&condition[7]);
         pthread_mutex_unlock(&mymutex[7]);
      }
   }
   return 0;
}

/** varcolorD stop function following jdec platform API schemas.
 *  @return integer stoping result.*/
int myvarcolorD_stop(){
   pthread_mutex_lock(&refmutex);
   if (varcolorD_refs>1){
      varcolorD_refs--;
      pthread_mutex_unlock(&refmutex);
   }
   else{
      varcolorD_refs=0;
      pthread_mutex_unlock(&refmutex);
      if((serve_color[7]==1)&&(color_active[7]==1)){
         color_active[7]=0;
         printf("varcolorD schema stop (simulated3D driver)\n");
         put_state(varcolorD_schema_id,slept);
         /* simulated3D thread goes sleep */
         pthread_mutex_lock(&mymutex[7]);
         state[7]=slept;
         pthread_mutex_unlock(&mymutex[7]);
      }
   }
   return 0;
}

/** simulated3D driver pthread function.*/
void *simulated3D_thread(void *id)
{
   int i,r;
   i=*((int*)id);

  printf("simulated3D driver started up\n", i);

 
  /* MAIN LOOP */
  while(simulated3D_terminate_command==0){
        
    pthread_mutex_lock(&mymutex[i]);

    if (state[i]==slept){
      printf("simulated3D thread in sleep mode\n");
      pthread_cond_wait(&condition[i],&mymutex[i]);
      printf("simulated3D thread woke up\n");
      pthread_mutex_unlock(&mymutex[i]);
      
    }else{
      
      pthread_mutex_unlock(&mymutex[i]);

      object_generator(i);

      if(serve_color[0]==1 && i==0){
		memcpy(colorA,ReadyColorSimA,width[0]*height[0]*3);
		speedcounter(colorA_schema_id);
		imageA_clock=lastimage;
      }
      else if(serve_color[1]==1 && i==1){
		memcpy(colorB,ReadyColorSimB,width[1]*height[1]*3);
		speedcounter(colorB_schema_id);
		imageB_clock=lastimage;
      }
      else if(serve_color[2]==1 && i==2){
		memcpy(colorC,ReadyColorSimC,width[2]*height[2]*3);
		speedcounter(colorC_schema_id);
		imageC_clock=lastimage;
      }
      else if(serve_color[3]==1 && i==3){
		memcpy(colorD,ReadyColorSimD,width[3]*height[3]*3);
		speedcounter(colorD_schema_id);
		imageD_clock=lastimage;
      }
      else if(serve_color[4]==1 && i==4){
		memcpy(myA.img,varReadyColorSimA,width[4]*height[4]*3);
		speedcounter(varcolorA_schema_id);
		myA.clock=lastimage;
      }
      else if(serve_color[5]==1 && i==5){
		memcpy(myB.img,varReadyColorSimB,width[5]*height[5]*3);
		speedcounter(varcolorB_schema_id);
		myB.clock=lastimage;
      }
      else if(serve_color[6]==1 && i==6){
		memcpy(myC.img,varReadyColorSimC,width[6]*height[6]*3);
		speedcounter(varcolorC_schema_id);
		myC.clock=lastimage;
      }
      else if(serve_color[7]==1 && i==7){
		memcpy(myD.img,varReadyColorSimD,width[7]*height[7]*3);
		speedcounter(varcolorD_schema_id);
		myD.clock=lastimage;
      }
      lastimage++;

    }// end else
  }// end while

  pthread_exit(0);
}

/** Determines if two 3D point are equals
 *  @param a point 1
 *  @param b point 2
 *  @return 1 if points are different, otherwise 0
 */
int different_pos(HPoint3D a,HPoint3D b){
	float difX=a.X-b.X;
	float difY=a.Y-b.Y;
	float difZ=a.Z-b.Z;

	if(difX==0&&difY==0&&difZ==0)
		return 0;
	else
		return 1;
}

/** Determines if an image configuration could be captured or not
 *  @param width width of the possible image
 *  @param height height of the possible image
 *  @return 1 if the size could be captured, otherwise 0
 */
int size_ok(int width, int height){
   if (width>0 && height>0)
      return 1;
   else
      return 0;
}

/** Determines if an image size from camera has the same size of an image
 *  @param c: camera
 *  @param i: image
 *  @return 1 if have the same size, otherwise 0
 */
int different_size(TPinHoleCamera c,int w, int h){
	return ((c.columns!=w) || (c.rows!=h));
}

/** simulated3D driver parse configuration file function.
 *  @param configfile path and name to the config file.
 *  @return 0 if parsing was successful or -1 if something went wrong.*/
int simulated3D_parseconf(char *configfile){

   int end_parse=0; int end_section=0; int driver_config_parsed=0;
   FILE *myfile;

   if ((myfile=fopen(configfile,"r"))==NULL){
      printf("simulated3D: cannot find config file\n");
      return -1;
   }

   do{
    
      char word[256],word2[256],buffer_file[1024];
      int i=0; int j=0;

      buffer_file[0]=fgetc(myfile);
    
      /* end of file */
      if (feof(myfile)){
         end_section=1;
         end_parse=1;
      
         /* line comment */
      }else if (buffer_file[0]=='#') {
         while(fgetc(myfile)!='\n');
      
         /* white spaces */
      }else if (buffer_file[0]==' ') {
         while(buffer_file[0]==' ') buffer_file[0]=fgetc(myfile);

         /* tabs */
      }else if(buffer_file[0]=='\t') {
         while(buffer_file[0]=='\t') buffer_file[0]=fgetc(myfile);
         /* storing line in buffer */
      }else{
      
         while(buffer_file[i]!='\n') buffer_file[++i]=fgetc(myfile);
         buffer_file[++i]='\0';

         if (i >= MAX_BUFFER-1) {
            printf("%s...\n", buffer_file);
            printf ("simulated3D: line too long in config file!\n");
            exit(-1);
         }
      
         /* first word of the line */
         if (sscanf(buffer_file,"%s",word)==1){
            if (strcmp(word,"driver")==0) {
               while((buffer_file[j]!='\n')&&(buffer_file[j]!=' ')&&(buffer_file[j]!='\0')&&(buffer_file[j]!='\t')) j++;
               sscanf(&buffer_file[j],"%s",word2);
	  
               /* checking if this section matchs our driver name */
               if (strcmp(word2,driver_name)==0){
                  /* the sections match */
                  do{
	      
                     char buffer_file2[1024],word3[256],word4[256];
                     char word5[256],word6[1024], word7[256],word8[256];
                     int k=0; int z=0;

                     buffer_file2[0]=fgetc(myfile);
	      
                     /* end of file */
                     if (feof(myfile)){
                        end_section=1;
                        end_parse=1;
		
                        /* line comment */
                     }else if (buffer_file2[0]=='#') {
                        while(fgetc(myfile)!='\n');
	      
                        /* white spaces */
                     }else if (buffer_file2[0]==' ') {
                        while(buffer_file2[0]==' ') buffer_file2[0]=fgetc(myfile);

                        /* tabs */
                     }else if(buffer_file2[0]=='\t') {
                        while(buffer_file2[0]=='\t') buffer_file2[0]=fgetc(myfile);

                        /* storing line in buffer */
                     }else{
		
                        while(buffer_file2[k]!='\n') buffer_file2[++k]=fgetc(myfile);
                        buffer_file2[++k]='\0';
		
                        /* first word of the line */
                        if (sscanf(buffer_file2,"%s",word3)==1){
                           if (strcmp(word3,"end_driver")==0) {
                              while((buffer_file2[z]!='\n')&&(buffer_file2[z]!=' ')&&(buffer_file2[z]!='\0')&&(buffer_file2[z]!='\t')) z++;
                              driver_config_parsed=1;
                              end_section=1;
                              end_parse=1;
		    
                           }else if (strcmp(word3,"driver")==0) {
                              while((buffer_file2[z]!='\n')&&
                                     (buffer_file2[z]!=' ')&&
                                     (buffer_file2[z]!='\0')&&
                                     (buffer_file2[z]!='\t')) z++;
                              printf("simulated3D: error in config file.\n'end_section' keyword required before starting new driver section.\n");
                              end_section=1; end_parse=1;

			   }else if(strcmp(word3,"configfile")==0){
			      int words;
                              while((buffer_file2[z]!='\n')&&
                                     (buffer_file2[z]!=' ')&&
                                     (buffer_file2[z]!='\0')&&
                                     (buffer_file2[z]!='\t')) z++;
                              words=sscanf(buffer_file2,"%s %s",word4,worldfile);
			      if(words<2){
				     printf("simulated3D: error in config file.\nYou must indicate world file path.\n");
				     end_section=1; end_parse=1;
                              }
                           }else if(strcmp(word3,"provides")==0){
                              int words;
                              while((buffer_file2[z]!='\n')&&
                                     (buffer_file2[z]!=' ')&&
                                     (buffer_file2[z]!='\0')&&
                                     (buffer_file2[z]!='\t')) z++;
                              words=sscanf(buffer_file2,"%s %s %s %s %s %s",
                                           word3,word4,word5,word6);
                              if(words==2){
                                 if(strcmp(word4,"colorA")==0){
                           	    serve_color[0]=1;
				    camCount++;
				    if (cameras[0].columns == -1){	
					   cameras[0].rows=SIFNTSC_ROWS;
					   cameras[0].columns=SIFNTSC_COLUMNS;
					   width[0]=SIFNTSC_COLUMNS;
					   height[0]=SIFNTSC_ROWS;
                                    }
                                    else{
                                          serve_color[0]=0;
                                          fprintf (stderr,"Camera %d already in use by an other color capture\n",camCount);
                                    }	
                                 }else if(strcmp(word4,"colorB")==0){
                         	    serve_color[1]=1;
				    camCount++;
				    if (cameras[1].columns == -1){	
					   cameras[1].rows=SIFNTSC_ROWS;
					   cameras[1].columns=SIFNTSC_COLUMNS;
					   width[1]=SIFNTSC_COLUMNS;
					   height[1]=SIFNTSC_ROWS;
                                    }
                                    else{
                                          serve_color[1]=0;
                                          fprintf (stderr,"Camera %d already in use by an other color capture\n",camCount);
                                    }	
                                 }else if(strcmp(word4,"colorC")==0){
        		            serve_color[2]=1;
				    camCount++;
				    if (cameras[2].columns == -1){	
					   cameras[2].rows=SIFNTSC_ROWS;
					   cameras[2].columns=SIFNTSC_COLUMNS;
					   width[2]=SIFNTSC_COLUMNS;
					   height[2]=SIFNTSC_ROWS;
                                    }
                                    else{
                                          serve_color[2]=0;
                                          fprintf (stderr,"Camera %d already in use by an other color capture\n",camCount);
                                    }	

                               }else if(strcmp(word4,"colorD")==0){
                         	    serve_color[3]=1;
				    camCount++;
				    if (cameras[3].columns == -1){	
					   cameras[3].rows=SIFNTSC_ROWS;
					   cameras[3].columns=SIFNTSC_COLUMNS;
					   width[3]=SIFNTSC_COLUMNS;
					   height[3]=SIFNTSC_ROWS;
                                    }
                                    else{
                                          serve_color[3]=0;
                                          fprintf (stderr,"Camera %d already in use by an other color capture\n",camCount);
                                    }
                                }
                              }
                              else if (words==4){
                                 if(strcmp(word4,"varcolorA")==0){
        		            serve_color[4]=1;
				    camCount++;
				    width[4]=atoi(word5);
				    height[4]=atoi(word6);
				    if (cameras[4].columns == -1){
					if(!size_ok(width[4],height[4])){
					   myA.height=cameras[4].rows=SIFNTSC_ROWS;
					   myA.width=cameras[4].columns=SIFNTSC_COLUMNS;
					}
					else{
					   myA.height=cameras[4].rows=height[4];
					   myA.width=cameras[4].columns=width[4];
					}
                                    }
                                    else{
                                          serve_color[4]=0;
                                          fprintf (stderr,"Camera %d already in use by an other color capture\n",camCount);
                                    }	                              
                                 }else if(strcmp(word4,"varcolorB")==0){
        		            serve_color[5]=1;
				    camCount++;
				    width[5]=atoi(word5);
				    height[5]=atoi(word6);
				    if (cameras[5].columns == -1){
					if(!size_ok(width[5],height[5])){
					   myB.height=cameras[5].rows=SIFNTSC_ROWS;
					   myB.width=cameras[5].columns=SIFNTSC_COLUMNS;
					}
					else{
					   myB.height=cameras[5].rows=height[5];
					   myB.width=cameras[5].columns=width[5];
					}
                                    }
                                    else{
                                          serve_color[5]=0;
                                          fprintf (stderr,"Camera %d already in use by an other color capture\n",camCount);
                                    }	
                                 }else if(strcmp(word4,"varcolorC")==0){
        		            serve_color[6]=1;
				    camCount++;
				    width[6]=atoi(word5);
				    height[6]=atoi(word6);
				    if (cameras[6].columns == -1){
					if(!size_ok(width[6],height[6])){
					   myC.height=cameras[6].rows=SIFNTSC_ROWS;
					   myC.width=cameras[6].columns=SIFNTSC_COLUMNS;
					}
					else{
					   myC.height=cameras[6].rows=height[6];
					   myC.width=cameras[6].columns=width[6];
					}
                                    }
                                    else{
                                          serve_color[6]=0;
                                          fprintf (stderr,"Camera %d already in use by an other color capture\n",camCount);
                                    }
                                 }else if(strcmp(word4,"varcolorD")==0){
                                    serve_color[7]=1;
				    camCount++;
				    width[7]=atoi(word5);
				    height[7]=atoi(word6);
				    if (cameras[7].columns == -1){
					if(!size_ok(width[7],height[7])){
					   myD.height=cameras[7].rows=SIFNTSC_ROWS;
					   myD.width=cameras[7].columns=SIFNTSC_COLUMNS;
					}
					else{
					   myD.height=cameras[7].rows=height[7];
					   myD.width=cameras[7].columns=width[7];
					}
                                    }
                                    else{
                                          serve_color[7]=0;
                                          fprintf (stderr,"Camera %d already in use by an other color capture\n",camCount);
                                    }
                                 }
                                }else{
                                   printf("simulated3D: provides line incorrect\n");
                                }
                           }else printf("simulated3D: i don't know what to do with '%s'\n",buffer_file2);
                        }
                     }
                  }while(end_section==0);
                  end_section=0;
               }
            }
         }
      }
   }while(end_parse==0);
  
   /* checking if a driver section was read */
   if(driver_config_parsed==1){
      if((serve_color[0]==0)&&(serve_color[1]==0)&&(serve_color[2]==0)&&(serve_color[3]==0)
	  &&(serve_color[4]==0)&&(serve_color[5]==0)&&(serve_color[6]==0)&&(serve_color[7]==0)){
         printf("simulated3D: warning! no color provided.\n");
      }
      return 0;
   }else return -1;
}

/** cleans up **/
void simulated3D_clean_up(){

}

/** simulated3D driver init function following jdec platform API for drivers.
 *  @param configfile path and name to the config file of this driver.*/
void simulated3D_init(char *configfile)
{
  int i;

  /* reseting serve color array and setting default options */
  for(i=0;i<MAXCAM;i++){
     serve_color[i]=0;
     color_active[i]=0;
     cameras[i].columns=-1;
  }

  /* we call the function to parse the config file */
  if(simulated3D_parseconf(configfile)==-1){
    printf("simulated3D: cannot initiate driver. configfile parsing error.\n");
    exit(-1);
  }

  for (i=0; i<MAXCAM; i++){
     if(simulated3D_thread_created[i]!=1){
        pthread_mutex_lock(&mymutex[i]);
        state[i]=slept;
        if (serve_color[i]){
           args[i]=i;
           pthread_create(&simulated3D_th[i],NULL,simulated3D_thread,(void*)&args[i]);
        }
        simulated3D_thread_created[i]=1;
        pthread_mutex_unlock(&mymutex[i]);
     }
  }

  /* Initialize the 3D objects from world file */
  if(tridimensional_objects_init()==-1){
     printf("world file %s does not exits\n",worldfile);
     exit(-1);
  }

  if(serve_color[0]==1){
      all[num_schemas].id = (int *) &colorA_schema_id;
      strcpy(all[num_schemas].name,"colorA");
      all[num_schemas].run = (runFn) mycolorA_run;
      all[num_schemas].stop = (stopFn) mycolorA_stop;
      printf("%s schema loaded (id %d)\n",all[num_schemas].name,num_schemas);
      (*(all[num_schemas].id)) = num_schemas;
      all[num_schemas].fps = 0.;
      all[num_schemas].k =0;
      all[num_schemas].state=slept;
      all[num_schemas].terminate = NULL;
      all[num_schemas].handle = NULL;
      num_schemas++;
      myexport("colorA","id",&colorA_schema_id);
      myexport("colorA","colorA",&colorA);
      myexport("colorA","clock", &imageA_clock);
      myexport("colorA","width",&width[0]);
      myexport("colorA","height",&height[0]);
      myexport("colorA","run",(void *)mycolorA_run);
      myexport("colorA","stop",(void *)mycolorA_stop);

      colorA=(char *)malloc (width[0]*height[0]*3);
      notReadyColorSimA=notReadyColorAA; ReadyColorSimA=readyColorAA;
  }

  if(serve_color[1]==1){ 
      all[num_schemas].id = (int *) &colorB_schema_id;
      strcpy(all[num_schemas].name,"colorB");
      all[num_schemas].run = (runFn) mycolorB_run;
      all[num_schemas].stop = (stopFn) mycolorB_stop;
      printf("%s schema loaded (id %d)\n",all[num_schemas].name,num_schemas);
      (*(all[num_schemas].id)) = num_schemas;
      all[num_schemas].fps = 0.;
      all[num_schemas].k =0;
      all[num_schemas].state=slept;
      all[num_schemas].terminate = NULL;
      all[num_schemas].handle = NULL;
      num_schemas++;
      myexport("colorB","id",&colorB_schema_id);
      myexport("colorB","colorB",&colorB);
      myexport("colorB","clock", &imageB_clock);
      myexport("colorB","width",&width[1]);
      myexport("colorB","height",&height[1]);
      myexport("colorB","run",(void *)mycolorB_run);
      myexport("colorB","stop",(void *)mycolorB_stop);

      colorB=(char *)malloc (width[1]*height[1]*3);
      notReadyColorSimB=notReadyColorBB; ReadyColorSimB=readyColorBB;
  }

  if(serve_color[2]==1){
      all[num_schemas].id = (int *) &colorC_schema_id;
      strcpy(all[num_schemas].name,"colorC");
      all[num_schemas].run = (runFn) mycolorC_run;
      all[num_schemas].stop = (stopFn) mycolorC_stop;
      printf("%s schema loaded (id %d)\n",all[num_schemas].name,num_schemas);
      (*(all[num_schemas].id)) = num_schemas;
      all[num_schemas].fps = 0.;
      all[num_schemas].k =0;
      all[num_schemas].state=slept;
      all[num_schemas].terminate = NULL;
      all[num_schemas].handle = NULL;
      num_schemas++;
      myexport("colorC","id",&colorC_schema_id);
      myexport("colorC","colorC",&colorC);
      myexport("colorC","clock", &imageC_clock);
      myexport("colorC","width",&width[2]);
      myexport("colorC","height",&height[2]);
      myexport("colorC","run",(void *)mycolorC_run);
      myexport("colorC","stop",(void *)mycolorC_stop);

      colorC=(char *)malloc (width[2]*height[2]*3);
      notReadyColorSimC=notReadyColorCC; ReadyColorSimC=readyColorCC; 
  }

  if(serve_color[3]==1){
      all[num_schemas].id = (int *) &colorD_schema_id;
      strcpy(all[num_schemas].name,"colorD");
      all[num_schemas].run = (runFn) mycolorD_run;
      all[num_schemas].stop = (stopFn) mycolorD_stop;
      printf("%s schema loaded (id %d)\n",all[num_schemas].name,num_schemas);
      (*(all[num_schemas].id)) = num_schemas;
      all[num_schemas].fps = 0.;
      all[num_schemas].k =0;
      all[num_schemas].state=slept;
      all[num_schemas].terminate = NULL;
      all[num_schemas].handle = NULL;
      num_schemas++;
      myexport("colorD","id",&colorD_schema_id);
      myexport("colorD","colorD",&colorD);
      myexport("colorD","clock", &imageD_clock);
      myexport("colorD","width",&width[3]);
      myexport("colorD","height",&height[3]);
      myexport("colorD","run",(void *)mycolorD_run);
      myexport("colorD","stop",(void *)mycolorD_stop);

      colorD=(char *)malloc (width[3]*height[3]*3);
      notReadyColorSimD=notReadyColorDD; ReadyColorSimD=readyColorDD;
  }
  /*creates new schema for varcolorA*/
  if(serve_color[4]==1){  
        all[num_schemas].id = (int *) &varcolorA_schema_id;
        strcpy(all[num_schemas].name,"varcolorA");
        all[num_schemas].run = (runFn) myvarcolorA_run;
        all[num_schemas].stop = (stopFn) myvarcolorA_stop;
        printf("%s schema loaded (id %d)\n",all[num_schemas].name,num_schemas);
        (*(all[num_schemas].id)) = num_schemas;
        all[num_schemas].fps = 0.;
        all[num_schemas].k =0;
        all[num_schemas].state=slept;
        all[num_schemas].terminate = NULL;
        all[num_schemas].handle = NULL;
        num_schemas++;
        myexport("varcolorA","id",&varcolorA_schema_id);
        myexport("varcolorA","varcolorA",&myA);
        myexport("varcolorA","run",(void *)myvarcolorA_run);
        myexport("varcolorA","stop",(void *)myvarcolorA_stop);
        
        myA.img=(char *)malloc (width[4]*height[4]*3);
	varReadyColorAA=(char *)malloc (width[4]*height[4]*3);
	varnotReadyColorAA=(char *)malloc (width[4]*height[4]*3);
	varnotReadyColorSimA=varnotReadyColorAA; varReadyColorSimA=varReadyColorAA;
  }
  /*creates new schema for varcolorB*/
  if(serve_color[5]==1){
        all[num_schemas].id = (int *) &varcolorB_schema_id;
        strcpy(all[num_schemas].name,"varcolorB");
        all[num_schemas].run = (runFn) myvarcolorB_run;
        all[num_schemas].stop = (stopFn) myvarcolorB_stop;
        printf("%s schema loaded (id %d)\n",all[num_schemas].name,num_schemas);
        (*(all[num_schemas].id)) = num_schemas;
        all[num_schemas].fps = 0.;
        all[num_schemas].k =0;
        all[num_schemas].state=slept;
        all[num_schemas].terminate = NULL;
        all[num_schemas].handle = NULL;
        num_schemas++;
        myexport("varcolorB","id",&varcolorB_schema_id);
        myexport("varcolorB","varcolorB",&myB);
        myexport("varcolorB","run",(void *)myvarcolorB_run);
        myexport("varcolorB","stop",(void *)myvarcolorB_stop);

        myB.img=(char *)malloc (width[5]*height[5]*3);
	varReadyColorBB=(char *)malloc (width[5]*height[5]*3);
	varnotReadyColorBB=(char *)malloc (width[5]*height[5]*3);
	varnotReadyColorSimB=varnotReadyColorBB; varReadyColorSimB=varReadyColorBB;
  }

  /*creates new schema for varcolorC*/
  if(serve_color[6]==1){
        all[num_schemas].id = (int *) &varcolorC_schema_id;
        strcpy(all[num_schemas].name,"varcolorC");
        all[num_schemas].run = (runFn) myvarcolorC_run;
        all[num_schemas].stop = (stopFn) myvarcolorC_stop;
        printf("%s schema loaded (id %d)\n",all[num_schemas].name,num_schemas);
        (*(all[num_schemas].id)) = num_schemas;
        all[num_schemas].fps = 0.;
        all[num_schemas].k =0;
        all[num_schemas].state=slept;
        all[num_schemas].terminate = NULL;
        all[num_schemas].handle = NULL;
        num_schemas++;
        myexport("varcolorC","id",&varcolorC_schema_id);
        myexport("varcolorC","varcolorC",&myC);
        myexport("varcolorC","run",(void *)myvarcolorC_run);
        myexport("varcolorC","stop",(void *)myvarcolorC_stop);

        myC.img=(char *)malloc (width[6]*height[6]*3);
	varReadyColorCC=(char *)malloc (width[6]*height[6]*3);
	varnotReadyColorCC=(char *)malloc (width[6]*height[6]*3);
	varnotReadyColorSimC=varnotReadyColorCC; varReadyColorSimC=varReadyColorCC;
  }

  /*creates new schema for varcolorD*/
  if(serve_color[7]==1){
        all[num_schemas].id = (int *) &varcolorD_schema_id;
        strcpy(all[num_schemas].name,"varcolorD");
        all[num_schemas].run = (runFn) myvarcolorD_run;
        all[num_schemas].stop = (stopFn) myvarcolorD_stop;
        printf("%s schema loaded (id %d)\n",all[num_schemas].name,num_schemas);
        (*(all[num_schemas].id)) = num_schemas;
        all[num_schemas].fps = 0.;
        all[num_schemas].k =0;
        all[num_schemas].state=slept;
        all[num_schemas].terminate = NULL;
        all[num_schemas].handle = NULL;
        num_schemas++;
        myexport("varcolorD","id",&varcolorD_schema_id);
        myexport("varcolorD","varcolorD",&myD);
        myexport("varcolorD","run",(void *)myvarcolorD_run);
        myexport("varcolorD","stop",(void *)myvarcolorD_stop);

        myD.img=(char *)malloc (width[7]*height[7]*3);
	varReadyColorDD=(char *)malloc (width[7]*height[7]*3);
	varnotReadyColorDD=(char *)malloc (width[7]*height[7]*3);
	varnotReadyColorSimD=varnotReadyColorDD; varReadyColorSimD=varReadyColorDD;
  }
  
  printf("simulated3D driver started up\n");
}
